﻿//-----------------------------------------------------------------------------
//
// Demonstrates stream callback functionality. For more stream details:
//
// SetStreamCallback Function:
//   https://support.labjack.com/docs/setstreamcallback-ljm-user-s-guide  
//
// Stream Functions:
//   https://support.labjack.com/docs/ljm-stream-functions-ljm-user-s-guide
//
// Stream Mode (datasheet):
//   https://support.labjack.com/docs/3-2-stream-mode-t-series-datasheet
//
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;

using LabJack;

namespace StreamCallback
{
    class Program
    {
        static void Main(string[] args)
        {
            Program prog = new Program();
            prog.performActions();
        }

        //Struct containing streaming information. Set with the stream settings.
        private struct StreamInfo
        {
            public StreamInfo(int handle, double scanRate, int scansPerRead,
                int numChannels, int[] scanList, string[] channelNames) : this()
            {
                this.handle = handle;
                this.scanRate = scanRate;
                this.scansPerRead = scansPerRead;
                this.numChannels = numChannels;
                
                this.scanList = new int[scanList.Length];
                Array.Copy(scanList, this.scanList, scanList.Length);
                
                this.channelNames = new string[channelNames.Length];
                Array.Copy(this.channelNames, channelNames, this.channelNames.Length);
                
                this.aDataSize = this.numChannels*this.scansPerRead;
                this.aData = new double[aDataSize];
                this.readCount = 0;
            }

            public int handle { get; set; }  //Device handle
            public double scanRate { get; set; }  //Stream scan rate
            public int scansPerRead { get; set; }  //Stream scans Per Read
            public int numChannels { get; set; }  //The number of channels streamed.
            public int[] scanList { get; set; }  //The scan list addresses.
            public string[] channelNames { get; set; }  //The scan list channel names.
            public int aDataSize { get; set; }  //The size of the stream data array. Calculated in constructor.
            public double[] aData { get; set; }  //The stream data array to store read data. Initialized in the constructor.
            public UInt64 readCount { get; set; }  //The number of stream reads performed. Updated in the callback.
        }

        private void showErrorMessage(LJM.LJMException e)
        {
            Console.Out.WriteLine("LJMException: " + e.ToString());
            Console.Out.WriteLine(e.StackTrace);
        }

        public void performActions()
        {
            int handle = 0;
            int devType = 0;
            int conType = 0;
            int serNum = 0;
            int ipAddr = 0;
            int port = 0;
            int maxBytesPerMB = 0;
            string ipAddrStr = "";
            string[] aNames;
            double[] aValues;
            int errorAddress = -1;

            //These variables are needed for the stream callback functionality.
            //To prevent memory issues, they cannot be garbage collected while
            //stream mode is running.
            LJM.StreamReadCallback callbackDelegate;  //The delegate for the callback function.
            IntPtr streamInfoPtr = IntPtr.Zero;  //Pointer to the user defined argument passed
                                                 //to the callback function. Defaulting to Zero, but
                                                 //will set later.

            try
            {
                //Open first found LabJack
                LJM.OpenS("ANY", "ANY", "ANY", ref handle);  // Any device, Any connection, Any identifier
                //LJM.OpenS("T8", "ANY", "ANY", ref handle);  // T8 device, Any connection, Any identifier
                //LJM.OpenS("T7", "ANY", "ANY", ref handle);  // T7 device, Any connection, Any identifier
                //LJM.OpenS("T4", "ANY", "ANY", ref handle);  // T4 device, Any connection, Any identifier
                //LJM.Open(LJM.CONSTANTS.dtANY, LJM.CONSTANTS.ctANY, "ANY", ref handle);  // Any device, Any connection, Any identifier

                LJM.GetHandleInfo(handle, ref devType, ref conType, ref serNum, ref ipAddr, ref port, ref maxBytesPerMB);
                LJM.NumberToIP(ipAddr, ref ipAddrStr);
                Console.WriteLine("Opened a LabJack with Device type: " + devType + ", Connection type: " + conType + ",");
                Console.WriteLine("Serial number: " + serNum + ", IP address: " + ipAddrStr + ", Port: " + port + ",");
                Console.WriteLine("Max bytes per MB: " + maxBytesPerMB);

                //Stream Configuration
                double scanRate = 2000;  //Scans per second
                int scansPerRead = (int)scanRate / 2;  //# scans returned by eStreamRead call
                const int numAddresses = 2;
                string[] aScanListNames = new String[] { "AIN0", "AIN1" };  //Scan list names to stream.
                int[] aTypes = new int[numAddresses];  //Dummy
                int[] aScanList = new int[numAddresses];  //Scan list addresses to stream. eStreamStart uses Modbus addresses.
                LJM.NamesToAddresses(numAddresses, aScanListNames, aScanList, aTypes);

                try
                {
                    //When streaming, negative channels and ranges can be configured for
                    //individual analog inputs, but the stream has only one settling time and
                    //resolution.

                    if (devType == LJM.CONSTANTS.dtT4)
                    {
                        //LabJack T4 configuration

                        //Stream settling is 0 (default) and stream resolution index is 0 (default).
                        aNames = new string[] {
                            "STREAM_SETTLING_US",
                            "STREAM_RESOLUTION_INDEX"
                        };
                        aValues = new double[] {
                            0,
                            0
                        };
                        LJM.eWriteNames(handle, aNames.Length, aNames, aValues, ref errorAddress);
                    }
                    else
                    {
                        //LabJack T7 and T8 configuration

                        // Settling and negative channel do not apply to the T8
                        if (devType == LJM.CONSTANTS.dtT7)
                        {
                            //All negative channels are single-ended
                            //Stream settling is 0 (default).
                            aNames = new string[] {
                                "AIN_ALL_NEGATIVE_CH",
                                "STREAM_SETTLING_US"
                            };
                            aValues = new double[] {
                                LJM.CONSTANTS.GND,
                                0
                            };
                            LJM.eWriteNames(handle, aNames.Length, aNames, aValues, ref errorAddress);
                        }

                        //Ensure triggered stream is disabled.
                        //Ensure internally-clocked stream.
                        //AIN0 and AIN1 ranges are set to ±10V (T7) or ±11V (T8).
                        //Stream resolution index is 0 (default).
                        aNames = new string[] {
                            "STREAM_TRIGGER_INDEX",
                            "STREAM_CLOCK_SOURCE",
                            "AIN0_RANGE",
                            "AIN1_RANGE",
                            "STREAM_RESOLUTION_INDEX"
                        };
                        aValues = new double[] {
                            0,
                            0,
                            10.0,
                            10.0,
                            0
                        };
                        LJM.eWriteNames(handle, aNames.Length, aNames, aValues, ref errorAddress);
                    }

                    int streamRuntime = 20000;

                    Console.WriteLine("\nStarting stream. Running for " + streamRuntime + " milliseconds.");

                    //Create a new StreamReadCallback delagate that uses the MyStreamReadCallback function.
                    callbackDelegate = new LJM.StreamReadCallback(MyStreamReadCallback);

                    //Create a StreamInfo struct, and get a pointer to it. This struct contains stream settings
                    //and will contain some data structures.
                    StreamInfo si = new StreamInfo(handle, scanRate, scansPerRead, numAddresses, aScanList, aScanListNames);
                    streamInfoPtr = Marshal.AllocHGlobal(Marshal.SizeOf(si));
                    Marshal.StructureToPtr(si, streamInfoPtr, false);

                    //Configure and start stream
                    LJM.eStreamStart(handle, scansPerRead, numAddresses, aScanList, ref scanRate);

                    //Set the stream callback. This needs to be done after starting the stream.
                    LJM.SetStreamCallback(handle, callbackDelegate, streamInfoPtr);

                    //Sleep in this thread while the callbacks run.
                    Thread.Sleep(streamRuntime);
                }
                catch (LJM.LJMException e)
                {
                    showErrorMessage(e);
                }
                Console.WriteLine("\nStopping stream.");
                LJM.eStreamStop(handle);
            }
            catch (LJM.LJMException e)
            {
                showErrorMessage(e);
            }

            LJM.CloseAll();  //Close all handles

            //Free the streamInfoPtr
            if (streamInfoPtr != IntPtr.Zero)
                Marshal.FreeHGlobal(streamInfoPtr);

            Console.WriteLine("\nDone.\nPress the enter key to exit.");
            Console.ReadLine();  //Pause for user
        }

        private void MyStreamReadCallback(IntPtr streamInfoPtr)
        {
            StreamInfo si;
            int devScanBacklog = 0;
            int ljmScanBacklog = 0;
            int skippedCur = 0;
            
            try
            {
                //Get our StreamInfo struct from the IntPtr.
                si = (StreamInfo)Marshal.PtrToStructure(streamInfoPtr, typeof(StreamInfo));

                //Read stream data
                LJM.eStreamRead(si.handle, si.aData, ref devScanBacklog, ref ljmScanBacklog);
                si.readCount++;

                //Count the skipped samples which are indicated by -9999 values. Missed
                //samples occur after a device's stream buffer overflows and are reported
                //after auto-recover mode ends.
                skippedCur = 0;
                foreach (double d in si.aData)
                {
                    if (d == -9999.00)
                        skippedCur++;
                }

                //Display current stream info.
                Console.WriteLine("\neStreamRead iteration " + si.readCount);
                Console.Write("  First scan out of " + si.scansPerRead + ": ");
                for (int j = 0; j < si.numChannels; j++)
                    Console.Write(si.channelNames[j] + " = " + si.aData[j].ToString("F4") + ", ");
                Console.WriteLine("\n  numSkippedScans: " + (skippedCur / si.numChannels) + ", deviceScanBacklog: " +
                    devScanBacklog + ", ljmScanBacklog: " + ljmScanBacklog);

                //Since we are incrementing readCount and want to maintain its count,
                //copy our struct data back into our pointer to keep updated struct info.
                //Note to set fDelete parameter to true to prevent memory leaks.
                Marshal.StructureToPtr(si, streamInfoPtr, true);
            }
            catch (LJM.LJMException e)
            {
                if(e.LJMError == LJM.LJMERROR.STREAM_NOT_RUNNING)
                    Console.WriteLine("Stream mode is not running.");
                else
                    showErrorMessage(e);
            }
            catch (Exception ex)
            {
                Console.WriteLine("Error in stream callback: " + ex.ToString());
            }
        }

    }
}
